#include <stdio.h>
#include <mpi.h>
#include <stdlib.h>
#include "boundary_conds_mpi.h"
#include "domain.h"

int apply_all_bcs(domain_t *d, int *bcs, int gridi, int gridj) {
	int xeven, yeven;

	xeven = 1 - (gridi % 2);
	yeven = 1 - (gridj % 2);

	/* note we're applying X BCs first, then y */

	if (xeven) {
		apply_bc(d, XRIGHT_DIR, bcs[1]);
		apply_bc(d, XLEFT_DIR,  bcs[0]);
	} else {
		apply_bc(d, XLEFT_DIR,  bcs[0]);
		apply_bc(d, XRIGHT_DIR, bcs[1]);
	}

	if (yeven) {
		apply_bc(d, YRIGHT_DIR, bcs[3]);
		apply_bc(d, YLEFT_DIR,  bcs[2]);
	} else {
		apply_bc(d, YLEFT_DIR,  bcs[2]);
		apply_bc(d, YRIGHT_DIR, bcs[3]);
	}

	return 0;
}

int apply_reflect_bc(domain_t *d, int dirn) {
	int i,j;
	int nx,   ny;
	int nxo, nyo;
	int nxt,  nyt;
	int nguard = d->Nguard;

	nx = d->Nx;
	ny = d->Ny;
	nxo= nx + nguard;
	nyo= ny + nguard;
	nxt= nx + 2*nguard;
	nyt= ny + 2*nguard;

	switch (dirn) {
		case XLEFT_DIR: 
			for (i=0; i<nguard; i++) {
				for (j=0; j<nyt; j++) {
					d->dens[i][j] =   d->dens[2*nguard-1-i][j];
					d->pres[i][j] =   d->pres[2*nguard-1-i][j];
					d->ener[i][j] =   d->ener[2*nguard-1-i][j];
					d->temp[i][j] =   d->temp[2*nguard-1-i][j];
					d->momx[i][j] = -(d->momx[2*nguard-1-i][j]);
					d->momy[i][j] =   d->momy[2*nguard-1-i][j];
				}
			}
			break;

		case XRIGHT_DIR:
			for (i=nxo; i<nxt; i++) {
				for (j=0; j<nyt; j++) {
					d->dens[i][j] =   d->dens[(2*nxo-1)-i][j];
					d->pres[i][j] =   d->pres[(2*nxo-1)-i][j];
					d->ener[i][j] =   d->ener[(2*nxo-1)-i][j];
					d->temp[i][j] =   d->temp[(2*nxo-1)-i][j];
					d->momx[i][j] = -(d->momx[(2*nxo-1)-i][j]);
					d->momy[i][j] =   d->momy[(2*nxo-1)-i][j];
				}
			}
			break;

		case YLEFT_DIR: 
			for (i=0; i<nxt; i++) {
				for (j=0; j<nguard; j++) {
					d->dens[i][j] =   d->dens[i][2*nguard-1-j];
					d->pres[i][j] =   d->pres[i][2*nguard-1-j];
					d->ener[i][j] =   d->ener[i][2*nguard-1-j];
					d->temp[i][j] =   d->temp[i][2*nguard-1-j];
					d->momx[i][j] =   d->momx[i][2*nguard-1-j];
					d->momy[i][j] = -(d->momy[i][2*nguard-1-j]);
				}
			}
			break;

		case YRIGHT_DIR: 
			for (i=0; i<nxt; i++) {
				for (j=nyo; j<nyt; j++) {
					d->dens[i][j] =   d->dens[i][(2*nyo-1)-j];
					d->pres[i][j] =   d->pres[i][(2*nyo-1)-j];
					d->ener[i][j] =   d->ener[i][(2*nyo-1)-j];
					d->temp[i][j] =   d->temp[i][(2*nyo-1)-j];
					d->momx[i][j] =   d->momx[i][(2*nyo-1)-j];
					d->momy[i][j] = -(d->momy[i][(2*nyo-1)-j]);
				}
			}
			break;
	}
	return 0;
}

int apply_periodic_bc(domain_t *d, int dirn) {
	int i,j;
	int nx,  ny;
	int nxo, nyo;
	int nxt, nyt;
	int nguard = d->Nguard;

	nx = d->Nx;
	ny = d->Ny;
	nxo= nx + nguard;
	nyo= ny + nguard;
	nxt= nx + 2*nguard;
	nyt= ny + 2*nguard;

	switch (dirn) {
		case XLEFT_DIR: 
			for (i=0; i<nguard; i++) {
				for (j=0; j<nyt; j++) {
					d->dens[i][j] = d->dens[i+nx][j];
					d->pres[i][j] = d->pres[i+nx][j];
					d->ener[i][j] = d->ener[i+nx][j];
					d->temp[i][j] = d->temp[i+nx][j];
					d->momx[i][j] = d->momx[i+nx][j];
					d->momy[i][j] = d->momy[i+nx][j];
				}
			}
			break;

		case XRIGHT_DIR:
			for (i=nxo; i<nxt; i++) {
				for (j=0; j<nyt; j++) {
					d->dens[i][j] = d->dens[i-nx][j];
					d->pres[i][j] = d->pres[i-nx][j];
					d->ener[i][j] = d->ener[i-nx][j];
					d->temp[i][j] = d->temp[i-nx][j];
					d->momx[i][j] = d->momx[i-nx][j];
					d->momy[i][j] = d->momy[i-nx][j];
				}
			}
			break;

		case YLEFT_DIR: 
			for (i=0; i<nxt; i++) {
				for (j=0; j<nguard; j++) {
					d->dens[i][j] = d->dens[i][j+ny];
					d->pres[i][j] = d->pres[i][j+ny];
					d->ener[i][j] = d->ener[i][j+ny];
					d->temp[i][j] = d->temp[i][j+ny];
					d->momx[i][j] = d->momx[i][j+ny];
					d->momy[i][j] = d->momy[i][j+ny];
				}
			}
			break;

		case YRIGHT_DIR: 
			for (i=0; i<nxt; i++) {
				for (j=nyo; j<nyt; j++) {
					d->dens[i][j] = d->dens[i][j-ny];
					d->pres[i][j] = d->pres[i][j-ny];
					d->ener[i][j] = d->ener[i][j-ny];
					d->temp[i][j] = d->temp[i][j-ny];
					d->momx[i][j] = d->momx[i][j-ny];
					d->momy[i][j] = d->momy[i][j-ny] ;
				}
			}
			break;
	}
	return 0;
}

int apply_zerograd_bc(domain_t *d, int dirn) {
	int i,j;
	int nx,  ny;
	int nxo,nyo;
	int nxt, nyt;
	int nguard = d->Nguard;

	nx = d->Nx;
	ny = d->Ny;
	nxo= nx + nguard;
	nyo= ny + nguard;
	nxt= nx + 2*nguard;
	nyt= ny + 2*nguard;

	switch (dirn) {
		case XLEFT_DIR: 
			for (i=0; i<nguard; i++) {
				for (j=0; j<nyt; j++) {
					d->dens[i][j] = d->dens[nguard+1][j];
					d->pres[i][j] = d->pres[nguard+1][j];
					d->ener[i][j] = d->ener[nguard+1][j];
					d->temp[i][j] = d->temp[nguard+1][j];
					d->momx[i][j] = d->momx[nguard+1][j];
					d->momy[i][j] = d->momy[nguard+1][j];
				}
			}
			break;

		case XRIGHT_DIR:
			for (i=nxo; i<nxt; i++) {
				for (j=0; j<nyt; j++) {
					d->dens[i][j] = d->dens[nxo-1][j];
					d->pres[i][j] = d->pres[nxo-1][j];
					d->ener[i][j] = d->ener[nxo-1][j];
					d->temp[i][j] = d->temp[nxo-1][j];
					d->momx[i][j] = d->momx[nxo-1][j];
					d->momy[i][j] = d->momy[nxo-1][j];
				}
			}
			break;

		case YLEFT_DIR: 
			for (i=0; i<nxt; i++) {
				for (j=0; j<nguard; j++) {
					d->dens[i][j] = d->dens[i][nguard+1];
					d->pres[i][j] = d->pres[i][nguard+1];
					d->ener[i][j] = d->ener[i][nguard+1];
					d->temp[i][j] = d->temp[i][nguard+1];
					d->momx[i][j] = d->momx[i][nguard+1];
					d->momy[i][j] = d->momy[i][nguard+1];
				}
			}
			break;

		case YRIGHT_DIR: 
			for (i=0; i<nxt; i++) {
				for (j=nyo; j<nyt; j++) {
					d->dens[i][j] = d->dens[i][nyo-1];
					d->pres[i][j] = d->pres[i][nyo-1];
					d->ener[i][j] = d->ener[i][nyo-1];
					d->temp[i][j] = d->temp[i][nyo-1];
					d->momx[i][j] = d->momx[i][nyo-1];
					d->momy[i][j] = d->momy[i][nyo-1] ;
				}
			}
			break;
	}
	return 0;
}

int apply_bc(domain_t *d, int dirn, int bc) {
	int size;	

	MPI_Comm_size(MPI_COMM_WORLD,&size);

	if ((dirn != XLEFT_DIR) && (dirn != XRIGHT_DIR) && (dirn != YLEFT_DIR) && (dirn != YRIGHT_DIR)) {
		fprintf(stderr,"apply_bc: invalid direction %d\n", dirn);
		return -1;
	}

	if (bc < 0) {
		switch (bc) {
			case BC_REFLECT: 
				apply_reflect_bc(d, dirn);
				break;

			case BC_PERIODIC: 
				apply_periodic_bc(d, dirn); 
				break;

			case BC_ZEROGRAD: 
				apply_zerograd_bc(d, dirn);
				break;

			default: fprintf(stderr,"apply_bc: invalid boundary condition %d\n", bc);
				 return -2;

		}
	} else {
		if (bc >= size) {
			fprintf(stderr,"apply_bc: invalid neighbor %d\n", bc);
			return -2;
		}

		return mpi_bc_vector(d,dirn,bc);
	}

	return 0;
}

int mpi_bc(domain_t *d, int dirn, int neigh) {
	/* first, ugly version which copies into a buffer */

	double *out, *in;
	int count;
	int nx, ny, ng;
	int i,j,ioff,joff;
	int ierr;
	int bufsize;
	int otherdirn;
	MPI_Status status;
	
	
	nx = d->Nx; ny = d->Ny; ng = d->Nguard;

	switch (dirn) {
		case XRIGHT_DIR: otherdirn = XLEFT_DIR;  break;
		case XLEFT_DIR:  otherdirn = XRIGHT_DIR; break;
		case YRIGHT_DIR: otherdirn = YLEFT_DIR;  break;
		case YLEFT_DIR:  otherdirn = YRIGHT_DIR; break;
	}

	if (dirn == XRIGHT_DIR || dirn == XLEFT_DIR) {
		bufsize = ng*(ny+2*ng)*NVARS;
		out = (double *)malloc(bufsize*sizeof(double));
		in  = (double *)malloc(bufsize*sizeof(double));
		if (out == NULL || in == NULL) {
			fprintf(stderr,"mpi_bc: Could not allocate memory.  Failing\n");
			return -1;
		}

		ioff = (dirn == XLEFT_DIR ? ng : nx);
		count = 0;
		for (i = ioff; i < ioff + ng; i++) {
			for (j=0; j<2*ng+ny; j++) {
				out[NVARS*count + DENSVAR] = d->dens[i][j];	
				out[NVARS*count + PRESVAR] = d->pres[i][j];	
				out[NVARS*count + ENERVAR] = d->ener[i][j];	
				out[NVARS*count + TEMPVAR] = d->temp[i][j];	
				out[NVARS*count + MOMXVAR] = d->momx[i][j];	
				out[NVARS*count + MOMYVAR] = d->momy[i][j];	
				count++; 
			}
		}
		ierr = MPI_Sendrecv((void *)out, bufsize, MPI_DOUBLE_PRECISION, neigh, dirn,
		                    (void *)in,  bufsize, MPI_DOUBLE_PRECISION, neigh, otherdirn, 
				    MPI_COMM_WORLD, &status);


		ioff = (dirn == XLEFT_DIR ? 0 : nx+ng);
		count = 0;
		for (i = ioff; i < ioff + ng; i++) {
			for (j=0; j<2*ng+ny; j++) {
				d->dens[i][j] = in[NVARS*count + DENSVAR];
				d->pres[i][j] =	in[NVARS*count + PRESVAR];
				d->ener[i][j] =	in[NVARS*count + ENERVAR];
				d->temp[i][j] =	in[NVARS*count + TEMPVAR];
				d->momx[i][j] =	in[NVARS*count + MOMXVAR];
				d->momy[i][j] =	in[NVARS*count + MOMYVAR];
				count++; 
			}
		}
	} else {
		bufsize = ng*(nx+2*ng)*NVARS;
		out = (double *)malloc(bufsize*sizeof(double));
		in  = (double *)malloc(bufsize*sizeof(double));
		if (out == NULL || in == NULL) {
			fprintf(stderr,"mpi_bc: Could not allocate memory.  Failing\n");
			return -1;
		}

		joff = (dirn == YLEFT_DIR ? ng : ny);
		count = 0;
		for (i =0; i < 2*ng+nx; i++) {
			for (j=joff; j<joff+ng; j++) {
				out[NVARS*count + DENSVAR] = d->dens[i][j];	
				out[NVARS*count + PRESVAR] = d->pres[i][j];	
				out[NVARS*count + ENERVAR] = d->ener[i][j];	
				out[NVARS*count + TEMPVAR] = d->temp[i][j];	
				out[NVARS*count + MOMXVAR] = d->momx[i][j];	
				out[NVARS*count + MOMYVAR] = d->momy[i][j];	
				count++; 
			}
		}
		ierr = MPI_Sendrecv((void *)out, bufsize, MPI_DOUBLE_PRECISION, neigh, dirn,
		                    (void *)in,  bufsize, MPI_DOUBLE_PRECISION, neigh, otherdirn, 
				    MPI_COMM_WORLD, &status);

		joff = (dirn == YLEFT_DIR ? 0 : ny+ng);
		count = 0;
		for (i = 0; i < 2*ng+nx; i++) {
			for (j=joff; j<joff+ng; j++) {
				d->dens[i][j] = in[NVARS*count + DENSVAR];
				d->pres[i][j] =	in[NVARS*count + PRESVAR];
				d->ener[i][j] =	in[NVARS*count + ENERVAR];
				d->temp[i][j] =	in[NVARS*count + TEMPVAR];
				d->momx[i][j] =	in[NVARS*count + MOMXVAR];
				d->momy[i][j] =	in[NVARS*count + MOMYVAR];
				count++; 
			}
		}
	}
	free(out); free(in);

	return 0;
}

/*
 * Uses MPI_Type commands to create a new type that copies straight out of the domain.
 * Faster, requires less memory + one less copy.
 */
int mpi_bc_vector(domain_t *d, int dirn, int neigh) {
	static MPI_Datatype xbctype;
	static MPI_Datatype ybctype;
	static int madetypes = 0;
	int nx = d->Nx;
	int ny = d->Ny;
	int ng = d->Nguard;
	int nxt, nyt;
	int ierr;
	MPI_Status status;
       
	nxt = 2*ng+nx;
	nyt = 2*ng+ny;

	if (!madetypes) {
		ierr = MPI_Type_vector(NVARS, nyt*ng, nxt*nyt, MPI_DOUBLE_PRECISION, &xbctype);
		ierr = MPI_Type_commit(&xbctype);

		ierr = MPI_Type_vector(nxt*NVARS, ng, nyt, MPI_DOUBLE_PRECISION, &ybctype);
		ierr = MPI_Type_commit(&ybctype);

		madetypes = 1;
	}


	switch (dirn) {
		case XRIGHT_DIR:
			ierr = MPI_Sendrecv(&(d->dens[nx][0])   , 1, xbctype, neigh, XRIGHT_DIR,
					    &(d->dens[nx+ng][0]), 1, xbctype, neigh, XLEFT_DIR, 
					    MPI_COMM_WORLD, &status);
			break;

		case XLEFT_DIR:
			ierr = MPI_Sendrecv(&(d->dens[ng][0]), 1, xbctype, neigh, XLEFT_DIR,
					    &(d->dens[ 0][0]), 1, xbctype, neigh, XRIGHT_DIR, 
					    MPI_COMM_WORLD, &status);
			break;

		case YRIGHT_DIR:
			ierr = MPI_Sendrecv(&(d->dens[0][ny]   ), 1, ybctype, neigh, YRIGHT_DIR,
					    &(d->dens[0][ny+ng]), 1, ybctype, neigh, YLEFT_DIR, 
					    MPI_COMM_WORLD, &status);
			break;

		case YLEFT_DIR:
			ierr = MPI_Sendrecv(&(d->dens[0][ng]), 1, ybctype, neigh, YLEFT_DIR,
					    &(d->dens[0][0] ), 1, ybctype, neigh, YRIGHT_DIR, 
					    MPI_COMM_WORLD, &status);
			break;
	}

	return 0;
}
