#include <math.h>
#include "eos.h"


double eos_gamma() { return EOS_GAMMA; }

/* 
 * forward_eos: given dens and ener, return pres and temp
 */
int forward_eos(int n, double *dens, double *ener, double *pres, double *temp) {

	int i;
	
	for (i=0; i<n; i++) {
		pres[i] = (EOS_GAMMA - 1.) * dens[i] * ener[i];
		temp[i] = ener[i] * ((EOS_GAMMA-1.)/EOS_GASCONST) * EOS_ABAR;
	}

	return 0;
}

int forward_eos_row(row_t *r) {
	int mustflip = 0;

	if (r->enerflag == FLAG_TOTALE) mustflip = 1;
	
	if (mustflip) ener_tot_to_internal(r);
	forward_eos(r->N, r->dens, r->ener, r->pres, r->temp);
	if (mustflip) ener_internal_to_tot(r);
	
	return 0;
}

int domain_forward_eos(domain_t *d) {

	int i,j;
	int nxt, nyt, nguard;
	
	nguard = d->Nguard;
	nxt = d->Nx+2*nguard;
	nyt = d->Ny+2*nguard;
	
	for (i=0; i<nxt; i++) {
		for (j=0; j<nyt; j++) {
			d->pres[i][j] = (EOS_GAMMA - 1.) * d->dens[i][j] * d->ener[i][j];
			d->temp[i][j] = d->ener[i][j] * ((EOS_GAMMA-1.)/EOS_GASCONST) * EOS_ABAR;
		}
	}

	return 0;
}

int domain_timestep(domain_t *d, double *dt) {

	int i,j;
	int nx, ny, ng;
	double mindt  = 1e+19;
	double dx, dy;
	double tdt;
	double vx, vy, cs;
	
	ng = d->Nguard;
	nx = d->Nx;
	ny = d->Ny;

	dx = d->x[1] - d->x[0]; // assuming uniform grid
	dy = d->y[1] - d->y[0]; // assuming uniform grid
	
	for (i=ng; i<nx+ng; i++) {
		for (j=ng; j<ny+ng; j++) {
			vx = fabs(d->momx[i][j]/d->dens[i][j]);
			vy = fabs(d->momy[i][j]/d->dens[i][j]);
			cs = sqrt(EOS_GAMMA*d->pres[i][j]/d->dens[i][j]);
			tdt = dx/(cs + vx);
			if (dy/(cs+vy) < tdt) tdt = dy/(cs+vy);
			if (tdt < mindt) mindt = tdt;
		}
	}
	*dt = mindt;

	return 0;
}

/*
 * backward_dp_eos: given dens and presure, return temp and ener
 */

int backward_dp_eos(int n, double *dens, double *pres, double *temp, double *ener) {

	int i;
	
	for (i=0; i<n; i++) {
		ener[i] = pres[i]/((EOS_GAMMA-1.)*dens[i]);
		temp[i] = ener[i] * ((EOS_GAMMA -1.)/EOS_GASCONST) * EOS_ABAR;
	}

	return 0;
}


int backward_dp_eos_row(row_t *r) {
	int mustflip = 0;

	if (r->enerflag == FLAG_TOTALE) mustflip = 1;
	
	if (mustflip) ener_tot_to_internal(r);
	backward_dp_eos(r->N, r->dens, r->pres, r->temp, r->ener);
	if (mustflip) ener_internal_to_tot(r);

	return 0;
}

int domain_backward_dp_eos(domain_t *d) {

	int i,j;
	int nx, ny, nguard;
	
	nguard = d->Nguard;
	nx = d->Nx+2*nguard;
	ny = d->Ny+2*nguard;
	
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			d->ener[i][j] = d->pres[i][j]/((EOS_GAMMA-1.)*d->dens[i][j]);
			d->temp[i][j] = d->ener[i][j] * ((EOS_GAMMA -1.)/EOS_GASCONST) * EOS_ABAR;
		}
	}

	return 0;
}
