#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "domain.h"
#include "kurganovlevy.h"
#include "eos.h"

int projectile_initvalues(domain_t *d, double psize, double pdens, double pvel) {

	double projcx = .2;
	double projcy = .5;
	double backgrounddens = 1.;
	double backgroundpres = 1.;

	int i, j;
	int nxt = d->Nx + 2*d->Nguard;
	int nyt = d->Ny + 2*d->Nguard;
	double r;
	double xx, yy, xx2;

	for (i=0; i<nxt; i++) {
		xx = d->x[i] - projcx;	
		xx2 = xx*xx;
		for (j=0; j<nyt; j++) {
			yy = d->y[j] - projcy;
			r = sqrt(xx2 + yy*yy);

			if (r < psize) {
				d->dens[i][j] = pdens - (pdens-backgrounddens)*(r/psize);
				d->momx[i][j] = pvel*d->dens[i][j];
			} else {
				d->dens[i][j] = backgrounddens;
				d->momx[i][j] = 0.;
			}
			d->pres[i][j] = backgroundpres;
			d->momy[i][j] = 0;
			d->temp[i][j] = 0;
			d->ener[i][j] = 0;
		}
	}

	domain_backward_dp_eos(d);

	return 0;
}

int sedov_initvalues(domain_t *d, double psize, double pener) {

	double cx = 0.;
	double cy = 0.;
	double backgrounddens = 1.;
	double backgroundpres = 1.e-5;

	int i, j;
	int nxt = d->Nx + 2*d->Nguard;
	int nyt = d->Ny + 2*d->Nguard;
	double r;
	double xx, yy, xx2;
	double pexp;

	pexp = pener*(eos_gamma() - 1.)/(3.14159*psize*psize) + backgroundpres;

	printf("Background pressure = %g\n", backgroundpres);
	printf("Explosion pressure = %g\n", pexp);

	for (i=0; i<nxt; i++) {
		xx = d->x[i] - cx;	
		xx2 = xx*xx;
		for (j=0; j<nyt; j++) {
			yy = d->y[j] - cy;
			r = sqrt(xx2 + yy*yy);

			if (r < psize) {
				d->pres[i][j] = pexp;
			} else {
				d->pres[i][j] = backgroundpres;
			}
			d->dens[i][j] = backgrounddens;
			d->momx[i][j] = 0;
			d->momy[i][j] = 0;
		}
	}

	domain_backward_dp_eos(d);

	return 0;
}
