#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "kurganovlevy.h"
#include "row.h"
#include "eos.h"
#include "reconstruction.h"
#include "boundary_conds.h"


int kl_timestep(domain_t *d, int *bcs, double dt, int gridi, int gridj) {

	/* strang splitting; XYYX */
	int i;
	row_t row, col;
	int nr, nc, ng;

	nr = d->Ny;
	nc = d->Nx;
	ng = d->Nguard;

	init_row(&row, nc+2*ng, d->x[0], d->x[nc+2*ng-1], d->enerflag);
	init_row(&col, nr+2*ng, d->y[0], d->y[nr+2*ng-1], d->enerflag);
			
	/* xsweep along all rows */
	for (i=ng; i<nr+ng; i++) {
		domain_extract_row(d, &row, i, XDIR);
		kl_row_rk2(&row, dt);
		domain_insert_row(d, &row, i, XDIR);
	}

	/* ysweep along all cols */
	for (i=ng; i<nc+ng; i++) {
		domain_extract_row(d, &col, i, YDIR);
		kl_row_rk2(&col, dt);
		domain_insert_row(d, &col, i, YDIR);
	}

	apply_all_bcs(d,bcs,gridi,gridj); 
	domain_forward_eos(d);

	for (i=ng; i<nc+ng; i++) {
		domain_extract_row(d, &col, i, YDIR);
		kl_row_rk2(&col, dt);
		domain_insert_row(d, &col, i, YDIR);
	}

	/* xsweep along all rows */
	for (i=ng; i<nr+ng; i++) {
		domain_extract_row(d, &row, i, XDIR);
		kl_row_rk2(&row, dt,gridi,gridj);
		domain_insert_row(d, &row, i, XDIR);
	}
	apply_all_bcs(d,bcs); 
	domain_forward_eos(d);

	free_row(&row);
	free_row(&col);

	return 0;
}
