#include <stdlib.h>
#include <stdio.h>
#include "row.h"

int init_row(row_t *r, int n, double startx, double endx, int enerflag) {
	
	int i;

	if (n < 0 || n > MAX_ROWSIZE) {
		fprintf(stderr,"init_row: Invalid x size of row %d\n", n);
		return -4;
	}

	r->N = n;

	r->x = (double *)malloc(n * sizeof(double));

	r->dens = (double *)malloc(n * sizeof(double));
	r->pres = (double *)malloc(n * sizeof(double));
	r->temp = (double *)malloc(n * sizeof(double));
	r->ener = (double *)malloc(n * sizeof(double));
	r->momx = (double *)malloc(n * sizeof(double));
	r->momy = (double *)malloc(n * sizeof(double));

	if (r->x == NULL || r->dens == NULL || r->pres == NULL || (r->temp) == NULL || r->ener == NULL) {
		fprintf(stderr,"init_row: Could not allocate memory for 1d arrays; what is going on?\n");
		free(r->x); free(r->dens); free(r->temp); free(r->ener); free(r->momx); free(r->momy);
		return -1;
	}

	for (i=0; i<n; i++) r->x[i] = startx + i*(endx-startx)/(n-1);

	r->enerflag = enerflag;

	r->alloced = (short int)1;
	return 0;
}

int free_row(row_t *r){

	if (r->alloced) {
		r->N = 0;	

		free(r->x); 
		free(r->dens);	
		free(r->pres);	
		free(r->ener);	
		free(r->temp);	
		free(r->momx);	
		free(r->momy);	

		r->alloced = (short int)0;
	}
	return 0;
}

int ener_tot_to_internal(row_t *r) {

	int i;
	double ke;

	if (r->enerflag == FLAG_SPECIFICINTERNALE) {
		fprintf(stderr,"Row is already internal energy; returning\n");
		return -1;
	}

	for (i=0; i<r->N; i++) {
		ke = 0.5*( (r->momx[i])*(r->momx[i]) + (r->momy[i])*(r->momy[i]) )/(r->dens[i]);
		r->ener[i] = (r->ener[i] - ke)/(r->dens[i]);
	}

	r->enerflag = FLAG_SPECIFICINTERNALE;

	return 0;
}

int ener_internal_to_tot(row_t *r) {

	int i;
	double ke;

	if (r->enerflag == FLAG_TOTALE) {
		fprintf(stderr,"Row is already total energy; returning\n");
		return -1;
	}

	for (i=0; i<r->N; i++) {
		ke = 0.5*( (r->momx[i])*(r->momx[i]) + (r->momy[i])*(r->momy[i]) )/(r->dens[i]);
		r->ener[i] = r->ener[i]*r->dens[i] + ke;
	}

	r->enerflag = FLAG_TOTALE;

	return 0;
}

int print_row(row_t *r, FILE *out) {
	
	int i;

	fprintf(out, "#x\tdens\tpres\tvelx\tvely\tener\ttemp\n");
	for (i=0; i<r->N; i++) {
		fprintf(out,"%g\t%g\t%g\t%g\t%g\t%g\t%g\n",r->x[i],r->dens[i],r->pres[i],r->momx[i]/r->dens[i],r->momy[i]/r->dens[i],r->ener[i],r->temp[i]);
	}
	return 0;

}
