#include <stdlib.h>
#include <stdio.h>
#include "row.h"
#include "domain.h"

int init_row(row_t *r, int n, double startx, double endx, int enerflag) {
	
	int i;
    double *tmp;

	if (n < 0 || n > MAX_ROWSIZE) {
		fprintf(stderr,"init_row: Invalid x size of row %d\n", n);
		return -4;
	}

	r->N = n;

	r->x = (double *)malloc(n * sizeof(double));
    r->data = (double **)malloc(NVARS*sizeof(double *));
    tmp = (double *)malloc(NVARS*n*sizeof(double));
    

	if (r->x == NULL || r->data == NULL || tmp == NULL) {
		fprintf(stderr,"init_row: Could not allocate memory for 1d arrays; what is going on?\n");
		free(r->x); free(r->data); free(tmp); 
		return -1;
	}

	r->data[DENSVAR] = &(tmp[DENSVAR*n]);
	r->data[ENERVAR] = &(tmp[ENERVAR*n]);
	r->data[PRESVAR] = &(tmp[PRESVAR*n]);
	r->data[TEMPVAR] = &(tmp[TEMPVAR*n]);
	r->data[MOMXVAR] = &(tmp[MOMXVAR*n]);
	r->data[MOMYVAR] = &(tmp[MOMYVAR*n]);

	for (i=0; i<n; i++) r->x[i] = startx + i*(endx-startx)/(n-1);

	r->enerflag = enerflag;

	r->alloced = (short int)1;
	return 0;
}

int free_row(row_t *r){

	if (r->alloced) {
		r->N = 0;	

		free(r->x); 
		free(r->data[DENSVAR]);	
		free(r->data);	

		r->alloced = (short int)0;
	}
	return 0;
}

int ener_tot_to_internal(row_t *r) {

	int i;
	double ke;

	if (r->enerflag == FLAG_SPECIFICINTERNALE) {
		fprintf(stderr,"Row is already internal energy; returning\n");
		return -1;
	}

	for (i=0; i<r->N; i++) {
		ke = 0.5*( (r->data[MOMXVAR][i])*(r->data[MOMXVAR][i]) + (r->data[MOMYVAR][i])*(r->data[MOMYVAR][i]) )/(r->data[DENSVAR][i]);
		r->data[ENERVAR][i] = (r->data[ENERVAR][i] - ke)/(r->data[DENSVAR][i]);
	}

	r->enerflag = FLAG_SPECIFICINTERNALE;

	return 0;
}

int ener_internal_to_tot(row_t *r) {

	int i;
	double ke;

	if (r->enerflag == FLAG_TOTALE) {
		fprintf(stderr,"Row is already total energy; returning\n");
		return -1;
	}

	for (i=0; i<r->N; i++) {
		ke = 0.5*( (r->data[MOMXVAR][i])*(r->data[MOMXVAR][i]) + (r->data[MOMYVAR][i])*(r->data[MOMYVAR][i]) )/(r->data[DENSVAR][i]);
		r->data[ENERVAR][i] = r->data[ENERVAR][i]*r->data[DENSVAR][i] + ke;
	}

	r->enerflag = FLAG_TOTALE;

	return 0;
}

int print_row(row_t *r, FILE *out) {
	
	int i;

	fprintf(out, "#x\tdens\tpres\tvelx\tvely\tener\ttemp\n");
	for (i=0; i<r->N; i++) {
		fprintf(out,"%g\t%g\t%g\t%g\t%g\t%g\t%g\n",r->x[i],r->data[DENSVAR][i],r->data[PRESVAR][i],r->data[MOMXVAR][i]/r->data[DENSVAR][i],r->data[MOMYVAR][i]/r->data[DENSVAR][i],r->data[ENERVAR][i],r->data[TEMPVAR][i]);
	}
	return 0;

}
