#include <mpi.h>
#include "communication.h"
#include "ljtypes.h"
#include "parameters.h"
 
long long pos_tag[3]    = { posX, posY, posZ };
long long neg_tag[3]    = { negX, negY, negZ };
int       n_send_tag[4] = { 0, 1, 4, 13 };
long long send_tag[13] 
= { posX,           posX|posY,      posY,         negX|posY,      posZ,  
    posX|posZ,      posX|posY|posZ, posY|posZ,    negX|posY|posZ, negX|posZ,
    negX|negY|posZ, negY|posZ,      posX|negY|posZ};

 
long long reverse_tag(long long tag) 
{
    return (tag&0x15)<<1 | (tag&0x2a)>>1;
}

 
void IdVector_pack(long long tag, int n_src, struct IdVector src[], int *n_dst, struct IdVector dst[])
{
    int i;
    for (i=0; i<n_src; i++) 
        if (src[i].tag & tag) {
            dst[*n_dst] = src[i];
            dst[*n_dst].route = src[i].route | tag;
            (*n_dst)++;
        }
} 

 
void Atom_pack(long long tag, int n_src, struct Atom src[], int *n_dst, struct IdVector dst[])
{
    int i;
    for (i=0; i<n_src; i++) 
        if (src[i].tag & tag) {
            dst[*n_dst].v[X] = src[i].q[X];
            dst[*n_dst].v[Y] = src[i].q[Y];
            dst[*n_dst].v[Z] = src[i].q[Z];
            dst[*n_dst].tag = src[i].tag;
            dst[*n_dst].route = tag;
            dst[*n_dst].localid = i;
            dst[*n_dst].id = src[i].id;
            (*n_dst)++; 
        }
}

 
void IdVector_prune(int *n, struct IdVector from[])
{
    int i = 0;
    while (i<*n && from[i].tag<to_be_removed) 
        ++i;

    while (i<*n) {
        from[i] = from[*n-1];
        --(*n);

        while (i<*n && from[i].tag<to_be_removed) 
            ++i;       
    }
}

 
void Atom_prune(int *n, struct Atom from[])
{
    int i = 0;
    while (i<*n && from[i].tag<to_be_removed) 
        ++i;

    while (i<*n) {
        from[i] = from[*n-1];
        --(*n);

        while (i<*n && from[i].tag<to_be_removed) 
            ++i;
    }
}

 
void IdVector_tagMove(long long tag, int *n_src, struct IdVector src[], int *n_dst, struct IdVector dst[])
{
    int i;
    for (i=0; i<*n_src; i++) {
        if (src[i].tag & tag) {
            dst[(*n_dst)++] = src[i];
            src[i].tag = to_be_removed;
        }
    }
}

 
void Atom_tagMove(long long tag, int *n_src, struct Atom src[], int *n_dst, struct Atom dst[])
{
    int i;
    for (i=0; i<*n_src; i++) {
        if (src[i].tag & tag) {
            dst[(*n_dst)++] = src[i];
            src[i].tag = to_be_removed;
        }
    }
}

 
void IdVector_move(long long tag, int *n_src, struct IdVector src[], int *n_dst, struct IdVector dst[])
{
    IdVector_tagMove(tag, n_src, src, n_dst, dst);
    IdVector_prune(n_src, src);
}

 
void IdVector_exchange(int *n_in, struct IdVector in[], int src, int n_out, struct IdVector out[], int dst,
                       int tag, int n_in_max, MPI_Comm comm)
{
    MPI_Status status;
    MPI_Sendrecv(out, n_out, MPI_IDVECTOR, dst, tag, in, n_in_max, MPI_IDVECTOR, src, tag, comm, &status);
    MPI_Get_count(&status, MPI_IDVECTOR, n_in);
}

 
void Atom_move(long long tag, int *n_src, struct Atom src[], int *n_dst, struct Atom dst[])
{
    Atom_tagMove(tag, n_src, src, n_dst, dst);
    Atom_prune(n_src, src);
}

 
void Atom_exchange(int *n_in, struct Atom in[], int src, int n_out, struct Atom out[], int dst,
                   int tag, int n_in_max, MPI_Comm comm)
{
    MPI_Status status;
    MPI_Sendrecv(out, n_out, MPI_ATOM, dst, tag, in, n_in_max, MPI_ATOM, src, tag, comm, &status);
    MPI_Get_count(&status, MPI_ATOM, n_in);
}

